from .body import Body


class Alter(Body):
    # 删除列
    def deleteColumn(self, table, liename):
        # ALTER TABLE tablename  DROP i;
        sql = '''ALTER TABLE {table}  DROP {lie}
        '''.format(table=self.getTablestr(table), lie=self.getLiestr(liename))
        return self.run(sql)

    # 修改列属性
    def setColumn(self, table, liename, newname, dataclass):
        sql = '''ALTER TABLE {table} 
                CHANGE {liename} {newname} {dataclass}
        '''.format(table=self.getTablestr(table),
                   liename=self.getLiestr(liename), newname=self.getLiestr(newname),
                   dataclass=dataclass)
        return self.run(sql)

    # 新增列
    def addColumn(self, table, liename, dataclass, other=""):
        # ALTER TABLE `tcl科技 (深证:000100)` add `昨日收盘` VARCHAR(255) AFTER `今日收盘`
        sql = '''ALTER TABLE {table}
                ADD {liename} {dataclass} 
                {other}
        '''.format(table=self.getTablestr(table),
                   liename=self.getLiestr(liename), dataclass=dataclass,
                   other=other)
        return self.run(sql)

    # 添加索引
    def addIndex(self, table, *index_lies, index_name=None):
        index_lie = ','.join(self.getLiestrs(index_lies))
        if not index_name:
            index_name = "_".join([table] + list(index_lies))
        self.run(f'create index {self.getLiestr(index_name)} on {self.getTablestr(table)}({index_lie});')

    # # 删除索引
    # def delIndex(self, table, index_name):
    #     self.run(f'DROP INDEX {self.getTablestr(index_name)}')



from base64 import b64encode, b64decode
from pyDes import PAD_PKCS5, CBC, des

__Des_Key0 = '%s&@W2*<FRW2'  # '%s+23das&@W2*<FRW2'
__Des_IV = b'\x52\x63\x78\x61\xBC\x48\x6A\x07'


def encrypt(user, passwd):
    k = des((__Des_Key0 % user)[:8], CBC, __Des_IV, padmode=PAD_PKCS5)
    return b64encode(k.encrypt(passwd)).decode('utf-8')


def decrypt(user, passwd_enc):
    k = des((__Des_Key0 % user)[:8], CBC, __Des_IV, padmode=PAD_PKCS5)
    return k.decrypt(b64decode(passwd_enc)).decode('utf-8')


class Body:
    def __init__(self, db_func, placeholder, class_map: dict):
        self._db_func = db_func
        self._db = db_func()
        self._cursor = self._db.cursor()

        self.placeholder = placeholder
        self.class_map = class_map

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()

    def commit(self):
        self._db.commit()

    def close(self):
        try:
            self._cursor.close()
            self._db.close()
        except:
            # traceback.print_exc()
            pass

    # 刷新连接
    def refresh(self):
        self.close()
        self._db = self._db_func()
        self._cursor = self._db.cursor()

    # 事务回滚
    def rollback(self):
        self._db.rollback()

    # 使用sql操作
    def run(self, sql):
        # print(sql)
        return self._cursor.execute(sql)

    def getLies(self) -> list:
        return [lc[0] for lc in self._cursor.description] \
            if self._cursor.description else []

    # 获取结果
    def getResult(self, data_class='dts') -> list:
        # 处理返回格式
        if data_class == 'ls':
            results = self._cursor.fetchall()
        else:
            lies = self.getLies()
            results = [dict(zip(lies, row)) for row in self._cursor.fetchall()]
        return results

    # 判断表是否存在
    def ifExist(self, table, def_schema):
        ls = table.split('.')
        if len(ls) == 1:
            schema = def_schema
        else:
            schema, table = ls
        self.run(f"SELECT 1 FROM INFORMATION_SCHEMA.TABLES WHERE TABLE_NAME = '{table}' and table_schema='{schema}'")
        return len(self._cursor.fetchall()) > 0

    def in_run(self, sql, *datas):
        try:
            self._cursor.executemany(sql, datas)
        except Exception as e:
            self.rollback()
            raise e

    def getTablestr(self, table) -> str:
        raise TypeError('需要实现方法-getTablestr')

    def getLiestr(self, lie) -> str:
        return f'"{lie.lower()}"'

    def getLiestrs(self, lies) -> list:
        return [self.getLiestr(lie) for lie in lies]



import os
import sqlite3
from .__sqltool__.action import Action
from .__sqltool__.body import encrypt, decrypt
from .__sqltool__.alter import Alter


class Sqlite(Action, Alter):
    def __init__(self, dbfilepath, if_new=True, ifencryp=False):
        
        # 打开数据库连接
        def db_func():
            if not if_new and not os.path.exists(dbfilepath):
                raise ValueError(f'没有 {dbfilepath} 文件!')
            db = sqlite3.connect(dbfilepath)
            return db
    
        class_map = {'varchar': 'text',
                     'text': 'text',
                     'bool': 'integer',
                     'int': 'integer',
                     'float': 'real'
                     }
        super().__init__(db_func, placeholder='?', class_map=class_map)

    def getTablestr(self, table):
        if '.' in table:
            t1, t2 = table.split('.')
            return f'"{t1}"."{t2}"'
        else:
            return f'"{table}"'

    # 判断表是否存在
    def ifExist(self, table):
        b = self.run(f"SELECT 1 FROM sqlite_master WHERE type='table' AND name='{table}'")
        return b and len(self._cursor.fetchall()) > 0



from .__sqltool__.action import Action
from .__sqltool__.body import encrypt, decrypt
from .__sqltool__.alter import Alter
import subprocess
try:
    from psycopg2 import connect as pgt_connect
    from psycopg2 import pool
except ModuleNotFoundError:
    subprocess.check_call(['pip','install', "psycopg2-binary"])
    from psycopg2 import connect as pgt_connect
    from psycopg2 import pool
    

class Pgsql(Action, Alter):
    def __init__(self, dbname, user, pwd,
                 host='127.0.0.1', port: int = 5432,
                 charset='UTF8',
                 ifencryp=True, custom_db_func = None):
        
        def db_func():
            db = pgt_connect(host=host, port=port,
                            user=user, password=decrypt(user, pwd) if ifencryp else pwd,
                            database=dbname)
            db.set_client_encoding(charset)
            return db
        
        class_map = {'varchar': 'varchar(255)',
                     'text': 'text',
                     'bool': 'bool',
                     'int': 'int8',
                     'float': 'float8'
                     }
        super().__init__(custom_db_func if custom_db_func else db_func, placeholder='%s', class_map=class_map)

    def getTablestr(self, table):
        if '.' in table:
            t1, t2 = table.split('.')
            return f'"{t1}"."{t2}"'
        else:
            return f'"{table}"'

    # 创建物化视图
    def createMaterializedView(self, view_name, run_sql):
        self.run(f'create materialized view "{view_name}" as {run_sql}')

    # 创建唯一约束
    pass

    # 刷新物化视图
    def refreshMaterializedView(self, view_name, if_lock=True):
        # 使用CONCURRENTLY的物化索引必须具有 unique 约束,否则会报错
        self.run(f'REFRESH MATERIALIZED VIEW {"" if if_lock else "CONCURRENTLY"} "{view_name}"')

    # 删除物化视图
    def deleteMaterializedView(self, view_name):
        self.run(f'drop materialized view if exists {view_name}')

    def createTable(self, table, lies: list,
                    colmap: dict = None,
                    key: list or str = None,
                    suffix: str = None,
                    parentable_valuetxt: tuple = None):
        if parentable_valuetxt:
            parent_table, valuetxt = parentable_valuetxt
            return self.createTable_child(table, parent_table, valuetxt)
        else:
            return Action.createTable(self, table, lies, suffix=suffix, colmap=colmap, key=key)

    def createTable_copy(self, table, copy_table):
        # 子表与父表结构保持一致
        sql = f'CREATE TABLE If Not Exists {self.getTablestr(table)} (LIKE {self.getTablestr(copy_table)} including all)'
        return self.run(sql)

    def createTable_parent(self, table, lies: list, key_lie: str, columnclassdict: dict = None, colmap: dict = None,
                           key=None, map_class='list', **kwargs):  # range
        # 分区父表
        return Action.createTable(self, table, lies, suffix=f"PARTITION BY {map_class} (\"{key_lie}\")",
                                  colmap=colmap, key=key, **kwargs)

    def createTable_child(self, table, parent_table, valuetxt):
        # 分区子表
        sql = f'CREATE TABLE If Not Exists {self.getTablestr(table)} PARTITION of {self.getTablestr(parent_table)} FOR values {valuetxt}'  # in ('123') from ('123') to ('125')
        return self.run(sql)

    # 解除子表分区关系
    def detachPartition(self, parentable, childtable):
        sql = f"ALTER TABLE {self.getTablestr(parentable)} detach PARTITION {self.getTablestr(childtable)}"
        return self.run(sql)

    # 绑定子表分区关系
    def attachPartition(self, parentable, childtable, valuetxt):
        sql = f"ALTER TABLE {self.getTablestr(parentable)} attach PARTITION {self.getTablestr(childtable)} FOR VALUES {valuetxt}"
        return self.run(sql)

    # 判断表是否存在
    def ifExist(self, table):
        return Action.ifExist(self, table, def_schema='public')

class PgsqlPool:
    def __init__(self, dbname, user, pwd, minconn=1, maxconn=10,
                 host='127.0.0.1', port: int = 5432,
                 charset='UTF8',
                 ifencryp=True, **kwargs):
        self._charset = charset
        self.pool = pool.SimpleConnectionPool(minconn, maxconn,
                                            host=host, port=port,
                                            user=user, password=decrypt(user, pwd) if ifencryp else pwd,
                                            database=dbname, **kwargs)
    
    def get(self) -> Pgsql:
        pgt = Pgsql(None, None, None, custom_db_func=self.pool.getconn)
        pgt._db.set_client_encoding(self._charset)
        pgt.close = lambda :self.pool.putconn(pgt._db)
        return pgt

    def close(self):
        try:
            self.pool.close()
        except:
            pass